///////////////////////////////////////////////////////////////////////////////
//
//	Simple example program demonstrating how to use a vertex shader using 
//	Cg. 
//
//	Gordon Wetzstein [wetzste1@cs.ubc.ca]
//	PSM Lab, University of British Columbia
//	March, 2009
//
///////////////////////////////////////////////////////////////////////////////
// includes

#include <stdio.h>
#include <iostream>
using namespace std;

// GL includes
#include <gl/glew.h>
#include <gl/glut.h>

// Cg includes
#include <cg/cg.h>
#include <cg/cggl.h>

///////////////////////////////////////////////////////////////////////////////
// GLUT callbacks and functions

// initialize glut window
void initGlut(int argc, char **argv);
// display callback
void displayFunc(void);
// idle callback
void idleFunc(void);
// reshape window callback
void reshapeFunc(int width, int height);
// mouse callback
void mouseFunc(int button, int state, int x, int y);
// mouse motion callback
void mouseMotionFunc(int x, int y);
// keyboard callback
void keyboardFunc(unsigned char key, int x, int y);
// callback for arrow keys
void specialFunc(int key, int x, int y);

///////////////////////////////////////////////////////////////////////////////
// Cg parameter

// Cg context
CGcontext	_cgContext;
//	Cg profile for the vertex shader
CGprofile	_cgVertexProfile;
//	Cg program for testing the performance
CGprogram	_cgProgram;
// parameter
CGparameter	_cgDeformation;

// Cg parameter for light position
CGparameter _cgLightPosition;
CGparameter _cgLightAmbient;
CGparameter _cgLightDiffuse;
CGparameter _cgLightSpecular;
CGparameter _cgLightAttenuation;

// Cg parameters for modelview and projection matrix
CGparameter _cgModelviewMatrix;
CGparameter _cgModelviewMatrixIT;
CGparameter _cgProjectionMatrix;
CGparameter _cgObjectMatrix;
CGparameter _cgObjectMatrixIT;

// Cg parameter for material 
CGparameter _cgMaterialEmission;
CGparameter _cgMaterialAmbient;
CGparameter _cgMaterialDiffuse;
CGparameter _cgMaterialSpecular;
CGparameter _cgMaterialShininess;

// CPU parameter
float		_deformation		= 0.0;
float		_deformationStep	= 0.1;
// CPU variable for light position
float		_lightPosition[]	= { 5.0, 15.0, 10.0, 1.0 };

// just indicates if shader is initialized
bool		_bShaderInitialized = false;

// load the shader 
bool		loadShader(void);
// destroy all Cg related stuff
void		destroyShader(void);

///////////////////////////////////////////////////////////////////////////////
// frame counter

void countFrames(void);
void renderBitmapString(float x, float y, float z, void *font, char *string);

// parameters for the framecounter
char pixelstring[256];
int cframe = 0;
int time = 0;
int timebase = 0;
float timestep = 0;

///////////////////////////////////////////////////////////////////////////////
// some other parameters

bool	bFullsreen = false;
int		nWindowID;

float	_teapotScale = 10.0;

///////////////////////////////////////////////////////////////////////////////
// parameters for the navigation

// camera attributes
float viewerPosition[3]		= { 0.0, 0.0, -50.0 };
float viewerDirection[3]	= { 0.0, 0.0, 0.0 };
float viewerUp[3]			= { 0.0, 1.0, 0.0 };

// rotation values for the navigation
float navigationRotation[3]	= { 0.0, 0.0, 0.0 };

// position of the mouse when pressed
int mousePressedX = 0, mousePressedY = 0;
float lastXOffset = 0.0, lastYOffset = 0.0, lastZOffset = 0.0;
// mouse button states
int leftMouseButtonActive = 0, middleMouseButtonActive = 0, rightMouseButtonActive = 0;
// modifier state
int shiftActive = 0, altActive = 0, ctrlActive = 0;

///////////////////////////////////////////////////////////////////////////////

void displayFunc(void) {

	GLenum err;

	// initialize shader
	if(!_bShaderInitialized) {

		_bShaderInitialized = loadShader();
		
		// check for errors
		err = glGetError();
		if(err != GL_NO_ERROR)
			printf("OpenGL ERROR after shader initialization: %s\n", gluErrorString(err));
	}

	//////////////////////////////////////////////////////////////////////////

	// clear the buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);

	// set some ambient material and light
	float materialAmbientSet[4];
	materialAmbientSet[0] = 0.3;
	materialAmbientSet[1] = 0.3;
	materialAmbientSet[2] = 0.3;
	materialAmbientSet[3] = 1.0;
	glMaterialfv(GL_FRONT, GL_AMBIENT, materialAmbientSet);
	glLightfv(GL_LIGHT0, GL_AMBIENT, materialAmbientSet);
	
	// set up projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, 1.33, 0.01, 100.0);
	
	glMatrixMode(GL_MODELVIEW);	
	glLoadIdentity();

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glLightfv(GL_LIGHT0, GL_POSITION, _lightPosition);

	// add mouse navigation 
	glTranslatef( viewerPosition[0], viewerPosition[1], viewerPosition[2] );
	glRotatef( navigationRotation[0], 1.0f, 0.0f, 0.0f );
	glRotatef( navigationRotation[1], 0.0f, 1.0f, 0.0f );

	// enable the shader and set all the parameters

	if(_bShaderInitialized) {

		// enable shader and set parameters
		cgGLEnableProfile( _cgVertexProfile );
		cgGLBindProgram( _cgProgram );

		// set matrix parameters
		cgGLSetStateMatrixParameter(_cgModelviewMatrix, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_IDENTITY);
		cgGLSetStateMatrixParameter(_cgModelviewMatrixIT, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_INVERSE_TRANSPOSE);
		cgGLSetStateMatrixParameter(_cgProjectionMatrix, CG_GL_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);

		// do some transformation that would be done in the glutSolidTeapot function
		// otherwise, need to include it in the modelview matrix
		glPushMatrix();
		glLoadIdentity();
		glRotatef(270.0, 1.0, 0.0, 0.0);
		glScalef(0.5*_teapotScale, 0.5*_teapotScale, 0.5*_teapotScale);
		glTranslatef(0.0, 0.0, -1.5);
		cgGLSetStateMatrixParameter(_cgObjectMatrix, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_IDENTITY);
		cgGLSetStateMatrixParameter(_cgObjectMatrixIT, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_INVERSE_TRANSPOSE);
		glPopMatrix();

		// set deformation parameter
		cgGLSetParameter1f(_cgDeformation, _deformation);

		// set light parameters
		cgGLSetParameter3f(_cgLightPosition, _lightPosition[0], _lightPosition[1], _lightPosition[2]);

		///////////////////////////////////////////////////////////////////////////////////////////////
		// set material parameters
		float materialEmission[4];
		glGetMaterialfv(GL_FRONT, GL_EMISSION, materialEmission);
		cgGLSetParameter3f(_cgMaterialEmission, materialEmission[0], materialEmission[1], materialEmission[2]);

		float materialAmbient[4];
		glGetMaterialfv(GL_FRONT, GL_AMBIENT, materialAmbient);
		cgGLSetParameter3f(_cgMaterialAmbient, materialAmbient[0], materialAmbient[1], materialAmbient[2]);

		float materialDiffuse[4];
		glGetMaterialfv(GL_FRONT, GL_DIFFUSE, materialDiffuse);
		cgGLSetParameter3f(_cgMaterialDiffuse, materialDiffuse[0], materialDiffuse[1], materialDiffuse[2]);

		float materialSpecular[4];
		glGetMaterialfv(GL_FRONT, GL_SPECULAR, materialSpecular);
		cgGLSetParameter3f(_cgMaterialSpecular, materialSpecular[0], materialSpecular[1], materialSpecular[2]);

		float materialShininess;
		glGetMaterialfv(GL_FRONT, GL_SHININESS, &materialShininess);
		cgGLSetParameter1f(_cgMaterialShininess, materialShininess);

		///////////////////////////////////////////////////////////////////////////////////////////////
		// set light parameters

		float lightAmbient[4];
		glGetLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
		cgGLSetParameter3f(_cgLightAmbient, lightAmbient[0], lightAmbient[1], lightAmbient[2]);
		
		float lightDiffuse[4];
		glGetLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
		cgGLSetParameter3f(_cgLightDiffuse, lightDiffuse[0], lightDiffuse[1], lightDiffuse[2]);
	
		float lightSpecular[4];
		glGetLightfv(GL_LIGHT0, GL_SPECULAR, lightSpecular);
		cgGLSetParameter3f(_cgLightSpecular, lightSpecular[0], lightSpecular[1], lightSpecular[2]);

		float lightAttenuation[3];
		glGetLightfv(GL_LIGHT0, GL_CONSTANT_ATTENUATION,	&(lightAttenuation[0]));
		glGetLightfv(GL_LIGHT0, GL_LINEAR_ATTENUATION,		&(lightAttenuation[1]));
		glGetLightfv(GL_LIGHT0, GL_QUADRATIC_ATTENUATION,	&(lightAttenuation[2]));
		cgGLSetParameter3f(_cgLightAttenuation, lightAttenuation[0], lightAttenuation[1], lightAttenuation[2]);

	}

	//glutSolidSphere(_teapotScale/2, 50, 50);
	glutSolidTeapot(_teapotScale);
	//glutSolidIcosahedron();

	// disable the shader 
	if(_bShaderInitialized) {
		// disable profile
		cgGLDisableProfile( _cgVertexProfile  );
	}

	// draw light source
	glDisable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glColor3f(1.0f, 1.0f, 1.0f);
	glTranslatef(_lightPosition[0], _lightPosition[1], _lightPosition[2]);
	glutSolidSphere(2.0, 50, 50);

	countFrames();
	glutSwapBuffers();
}

///////////////////////////////////////////////////////////////////////////////

void initGlut(int argc, char **argv) {

	// GLUT Window Initialization:
	glutInit (&argc, argv);
	glutInitWindowSize (640, 480);
	glutInitWindowPosition(100, 100);
	glutInitDisplayMode ( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	nWindowID = glutCreateWindow ("CS 314 CgExample Vertex Shader");

	// Register callbacks:
	glutDisplayFunc		(displayFunc);
	glutReshapeFunc		(reshapeFunc);
	glutKeyboardFunc	(keyboardFunc);
	glutSpecialFunc		(specialFunc);
	glutMouseFunc		(mouseFunc);
	glutMotionFunc		(mouseMotionFunc);
	glutIdleFunc		(idleFunc);
}

///////////////////////////////////////////////////////////////////////////////

void idleFunc(void) {
	glutPostRedisplay();
}

///////////////////////////////////////////////////////////////////////////////

void reshapeFunc(int width, int height) {
	glViewport(0, 0, width, height);
}

///////////////////////////////////////////////////////////////////////////////

// mouse callback
void mouseFunc(int button, int state, int x, int y) {
	
	// get the mouse buttons
	if (button == GLUT_LEFT_BUTTON)
		if (state == GLUT_DOWN) {
			leftMouseButtonActive += 1;
		} else
			leftMouseButtonActive -= 1;
	else if (button == GLUT_MIDDLE_BUTTON)
		if (state == GLUT_DOWN) {
			middleMouseButtonActive += 1;
			lastXOffset = 0.0;
			lastYOffset = 0.0;
		} else
			middleMouseButtonActive -= 1;
	else if (button == GLUT_RIGHT_BUTTON)
		if (state == GLUT_DOWN) {
			rightMouseButtonActive += 1;
			lastZOffset = 0.0;
		} else
			rightMouseButtonActive -= 1;

	mousePressedX = x;
	mousePressedY = y;
}

///////////////////////////////////////////////////////////////////////////////

void mouseMotionFunc(int x, int y) {
	
	float xOffset = 0.0, yOffset = 0.0, zOffset = 0.0;
	
	// rotatation
	if (leftMouseButtonActive) {

		navigationRotation[0] += ((mousePressedY - y) * 180.0f) / 200.0f;
		navigationRotation[1] += ((mousePressedX - x) * 180.0f) / 200.0f;

		mousePressedY = y;
		mousePressedX = x;

	}
	// panning
	else if (middleMouseButtonActive) {

		xOffset = (mousePressedX + x);
		if (!lastXOffset == 0.0) {
			viewerPosition[0]	-= (xOffset - lastXOffset) / 8.0;
			viewerDirection[0]	-= (xOffset - lastXOffset) / 8.0;
		}
		lastXOffset = xOffset;

		yOffset = (mousePressedY + y);
		if (!lastYOffset == 0.0) {
			viewerPosition[1]	+= (yOffset - lastYOffset) / 8.0;
			viewerDirection[1]	+= (yOffset - lastYOffset) / 8.0;	
		}	
		lastYOffset = yOffset;

	}
	// depth movement
	else if (rightMouseButtonActive) {
		zOffset = (mousePressedX + x);
		if (!lastZOffset == 0.0) {
			viewerPosition[2] -= (zOffset - lastZOffset) / 5.0;
			viewerDirection[2] -= (zOffset - lastZOffset) / 5.0;
		}
		lastZOffset = zOffset;
	}
}

///////////////////////////////////////////////////////////////////////////////

void keyboardFunc(unsigned char key, int x, int y) {
    
	switch(key) {
		
		////////////////////////////////////////////

#ifdef WIN32
		// exit on escape
		case '\033':
			destroyShader();
			exit(0);
			break;
#endif

		////////////////////////////////////////////
			
		// switch to fullscreen
		case 'f':

			bFullsreen = !bFullsreen;
			if(bFullsreen) 
				glutFullScreen();
			else {
				glutSetWindow(nWindowID);
				glutPositionWindow(100, 100);
				glutReshapeWindow(640, 480);
			}
			break;

		////////////////////////////////////////////
	}
}

///////////////////////////////////////////////////////////////////////////////

void specialFunc(int key, int x, int y) {
	if(key == GLUT_KEY_UP) {
		_deformation += _deformationStep;
	} else if(key == GLUT_KEY_DOWN) {
		_deformation = ((_deformation-_deformationStep) >= 0.0) ? _deformation-_deformationStep : 0.0;
	}
}

///////////////////////////////////////////////////////////////////////////////

void countFrames(void)  {

	time=glutGet(GLUT_ELAPSED_TIME);
	cframe++;
	if (time - timebase > 50) {
		sprintf(pixelstring, "deformation: %4.1f | fps: %4.2f", _deformation, cframe*1000.0/(time-timebase));		
		timebase = time;
		cframe = 0;	
	} 
	glPushMatrix();
	glLoadIdentity();
	glDisable(GL_LIGHTING);
	glColor4f(1.0,1.0,1.0,1.0);
	glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		gluOrtho2D(0, 200, 0, 200);
	glMatrixMode(GL_MODELVIEW);

	// render the string
	renderBitmapString(5,5,0.0,GLUT_BITMAP_HELVETICA_10,pixelstring);
		
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

///////////////////////////////////////////////////////////////////////////////

void renderBitmapString(float x, float y, float z, void *font, char *string) {
  char *c;
  glRasterPos3f(x, y,z);
  for (c=string; *c != '\0'; c++) {
    glutBitmapCharacter(font, *c);
  }
}

///////////////////////////////////////////////////////////////////////////////

bool	
loadShader(void) {

	/////////////////////////////////////////////////////////////////////
	// init fragment shader profile

	_cgVertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);   

	/////////////////////////////////////////////////////////////////////
	// init context

	if(!_cgContext)
		_cgContext = cgCreateContext();

	/////////////////////////////////////////////////////////////////////
	//	init shader

	_cgProgram = cgCreateProgramFromFile(	_cgContext, 
											CG_SOURCE, 
											"VertexShader.cg", 
											_cgVertexProfile, 
											NULL, 
											NULL );

	cgGLLoadProgram( _cgProgram );

	// get all the parameters
	_cgDeformation			= cgGetNamedParameter(_cgProgram, "deformation");
	// light parameters
	_cgLightPosition		= cgGetNamedParameter(_cgProgram, "lightPosition");
	_cgLightAmbient			= cgGetNamedParameter(_cgProgram, "lightAmbient");
	_cgLightDiffuse			= cgGetNamedParameter(_cgProgram, "lightDiffuse");
	_cgLightSpecular		= cgGetNamedParameter(_cgProgram, "lightSpecular");
	_cgLightAttenuation		= cgGetNamedParameter(_cgProgram, "lightAttenuation");
	// matrix parameters
	_cgModelviewMatrix		= cgGetNamedParameter(_cgProgram, "modelViewMatrix");
	_cgModelviewMatrixIT	= cgGetNamedParameter(_cgProgram, "modelViewMatrixIT");
	_cgProjectionMatrix		= cgGetNamedParameter(_cgProgram, "projectionMatrix");
	_cgObjectMatrix			= cgGetNamedParameter(_cgProgram, "objectMatrix");
	_cgObjectMatrixIT		= cgGetNamedParameter(_cgProgram, "objectMatrixIT");
	// material parameters
	_cgMaterialEmission		= cgGetNamedParameter(_cgProgram, "materialEmission");
	_cgMaterialAmbient		= cgGetNamedParameter(_cgProgram, "materialAmbient");
	_cgMaterialDiffuse		= cgGetNamedParameter(_cgProgram, "materialDiffuse");
	_cgMaterialSpecular		= cgGetNamedParameter(_cgProgram, "materialSpecular");
	_cgMaterialShininess	= cgGetNamedParameter(_cgProgram, "materialShininess");

	/////////////////////////////////////////////////////////////////////
	// check for errors

	GLenum error = glGetError();
	if(error != GL_NO_ERROR) {
		cout << "GL ERROR: " << gluErrorString(error) << endl;
		return false;
	}

	if(CGerror cgError = cgGetError()) {
		cout << "Cg ERROR: " << cgGetErrorString(cgError) << endl;
		return false;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////

void		
destroyShader(void) {

	if(_bShaderInitialized) {
		cgDestroyProgram(_cgProgram);
	}
}

///////////////////////////////////////////////////////////////////////////////

void main(int argc, char **argv) {
	
	initGlut(argc, argv);

	cout << "Simple Cg vertex shader example." << endl;
	cout << "Gordon Wetzstein [wetzste1@cs.ubc.ca]" << endl << endl;

	cout << "keys:	Esc		- terminate program" << endl;
	cout << "	arrow up	- increase deform" << endl;
	cout << "	arrow down	- decrease deform" << endl << endl;
	cout << "	Mouse Left	- rotation" << endl;
	cout << "	Mouse Right	- zoom" << endl << endl;

	cout << "OpenGL version: " << glGetString(GL_VERSION) << endl;
	cout << "OpenGL vendor: " << glGetString(GL_VENDOR) << endl;
	cout << "OpenGL renderer: " << glGetString(GL_RENDERER) << endl << endl;

	// hand over to glut
	glutMainLoop();
		
}



